from fastapi import FastAPI, Depends, HTTPException, status
from fastapi.security import HTTPBearer, HTTPAuthorizationCredentials
from sklearn.preprocessing import PolynomialFeatures
from sentence_transformers import SentenceTransformer
from fastapi.middleware.cors import CORSMiddleware
from scipy.interpolate import interp1d
from pydantic import BaseModel
from typing import List
import numpy as np
import uvicorn
import tiktoken
import torch
import os

# 环境变量传入
sk_key = os.environ.get('sk-key', 'sk-nvO0YlZEIrljq0QXB35b7e4f711d4087Ac215c18De75A407')
is_ver = False

# 创建FastAPI的实例
app = FastAPI()

# 设置FastAPI中间件
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# 创建一个HTTPBearer实例
security = HTTPBearer()
print("\n")

# 检测是否有GPU可用，如果有则使用cuda设备，否则使用cpu设备
device = str(torch.device('cuda' if torch.cuda.is_available() else 'cpu'))
if torch.cuda.is_available():
    print("===========================")
    print('本次加载模型的设备为GPU: ', torch.cuda.get_device_name(0))
    print("===========================")
else:
    print("===========================")
    print('本次加载模型的设备为CPU.')
    print("===========================")

# 预加载模型
model = SentenceTransformer('./models/m3e-large', device=device)


class EmbeddingRequest(BaseModel):
    input: List[str]
    model: str


class EmbeddingResponse(BaseModel):
    data: list
    model: str
    object: str
    usage: dict


def num_tokens_from_string(string: str) -> int:
    """返回文本字符串中的标记数"""
    encoding = tiktoken.get_encoding('cl100k_base')
    num_tokens = len(encoding.encode(string))
    return num_tokens


def interpolate_vector(vector, target_length):
    """插值向量"""
    original_indices = np.arange(len(vector))
    target_indices = np.linspace(0, len(vector) - 1, target_length)
    f = interp1d(original_indices, vector, kind='linear')
    return f(target_indices)


def expand_features(embedding, target_length):
    """扩展功能"""
    poly = PolynomialFeatures(degree=2)
    expanded_embedding = poly.fit_transform(embedding.reshape(1, -1))
    expanded_embedding = expanded_embedding.flatten()
    if len(expanded_embedding) > target_length:
        # 如果扩展后的特征超过目标长度，可以通过截断或其他方法来减少维度
        expanded_embedding = expanded_embedding[:target_length]
    elif len(expanded_embedding) < target_length:
        # 如果扩展后的特征少于目标长度，可以通过填充或其他方法来增加维度
        expanded_embedding = np.pad(expanded_embedding, (0, target_length - len(expanded_embedding)))
    return expanded_embedding


@app.post("/v1/embeddings", response_model=EmbeddingResponse)
async def get_embeddings(request: EmbeddingRequest, credentials: HTTPAuthorizationCredentials = Depends(security)):
    """调用接口"""
    if is_ver and credentials.credentials != sk_key:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid Authorization code",
        )

    # 计算嵌入向量和tokens数量 
    embeddings = [model.encode(text) for text in request.input]

    # 如果嵌入向量的维度不为1536，则使用插值法扩展至1536维度
    # embeddings = [interpolate_vector(embedding, 1536) if len(embedding) < 1536 else embedding for embedding in embeddings]

    # 如果嵌入向量的维度不为1536，则使用特征扩展法扩展至1536维度
    embeddings = [expand_features(embedding, 1536) if len(embedding) < 1536 else embedding for embedding in embeddings]

    # Min-Max normalization embeddings = [(embedding - np.min(embedding)) / (np.max(embedding) - np.min(embedding)) if np.max(embedding) != np.min(embedding) else embedding for embedding in embeddings]
    embeddings = [embedding / np.linalg.norm(embedding) for embedding in embeddings]

    # 将Numpy数组转换为列表
    embeddings = [embedding.tolist() for embedding in embeddings]
    prompt_tokens = sum(len(text.split()) for text in request.input)
    total_tokens = sum(num_tokens_from_string(text) for text in request.input)

    response = {
        "data": [
            {
                "embedding": embedding,
                "index": index,
                "object": "embedding"
            } for index, embedding in enumerate(embeddings)
        ],
        "model": request.model,
        "object": "list",
        "usage": {
            "prompt_tokens": prompt_tokens,
            "total_tokens": total_tokens,
        }
    }

    return response


if __name__ == "__main__":
    uvicorn.run("main:app", host='0.0.0.0', port=9200, workers=1)
